#include "stdafx.h"
#include "mesh.h"

DWORD MeshVertex::decl[] = 
{ 
	D3DVSD_STREAM(0), 
	D3DVSD_REG(0, D3DVSDT_FLOAT3), // position
	D3DVSD_REG(1, D3DVSDT_FLOAT3), // normal
	D3DVSD_REG(2, D3DVSDT_FLOAT2), // texcoord
	D3DVSD_END() 
};

oVector3 GetBoxGeomLength(const dGeomID& geom) throw()
{
	dVector3 length;
	dGeomBoxGetLengths(geom, length);
	oVector3 temp;
	temp.CreateFromdVector3(length);
	return temp;
}

inline void Mesh::Allocate(const ushort& numVertices_, const ushort& numIndices_, const ushort& numPrimitives_)
{
	numVertices = numVertices_;
    numIndices = numIndices_;
	numPrimitives = numPrimitives_;
	oAlloc(vertices, numVertices);
	oAlloc(indices, numIndices);
	oAlloc(primitives, numPrimitives);
}

inline void Mesh::CreateBoxFace(MeshVertex* const& buffer, const oVector3& normal, const oReal& u, const oReal& v) throw()
{
	buffer[1].t = oVector2(_R(0), _R(0));
	buffer[0].t = oVector2(_R(0), v);
	buffer[3].t = oVector2(u, v);
	buffer[2].t = oVector2(u, _R(0));
	buffer[0].n = buffer[1].n = buffer[2].n = buffer[3].n = normal;
}

Mesh::Mesh() : numVertices(0), numIndices(0), numPrimitives(0) {}

void Mesh::CreateBox(const oVector3& length)
{
	CreateBox(length, length);
}

void Mesh::CreateBox(const oVector3& length, const oVector3 textureScale)
{
	// allocate
	Allocate(24, 36, 1);
	// fill vertices
    MeshVertex* buffer = vertices;
	buffer[0].p = oVector3(_R(-0.5), _R(-0.5), _R(-0.5)); // front
	buffer[1].p = oVector3(_R(-0.5), _R( 0.5), _R(-0.5));
	buffer[2].p = oVector3(_R( 0.5), _R( 0.5), _R(-0.5));
	buffer[3].p = oVector3(_R( 0.5), _R(-0.5), _R(-0.5));
	CreateBoxFace(buffer, oVector3(_R(0), _R(0), _R(-1)), length.x / textureScale.x, length.y / textureScale.y); buffer += 4;
	buffer[0].p = oVector3(_R( 0.5), _R(-0.5), _R( 0.5)); // back
	buffer[1].p = oVector3(_R( 0.5), _R( 0.5), _R( 0.5));
	buffer[2].p = oVector3(_R(-0.5), _R( 0.5), _R( 0.5));
	buffer[3].p = oVector3(_R(-0.5), _R(-0.5), _R( 0.5));
	CreateBoxFace(buffer, oVector3(_R(0), _R(0), _R( 1)), length.x / textureScale.x, length.y / textureScale.y); buffer += 4;
	buffer[0].p = oVector3(_R(-0.5), _R(-0.5), _R( 0.5)); // left
	buffer[1].p = oVector3(_R(-0.5), _R( 0.5), _R( 0.5));
	buffer[2].p = oVector3(_R(-0.5), _R( 0.5), _R(-0.5));
	buffer[3].p = oVector3(_R(-0.5), _R(-0.5), _R(-0.5));
	CreateBoxFace(buffer, oVector3(_R(-1), _R(0), _R(0)), length.z / textureScale.z, length.y / textureScale.y); buffer += 4;
	buffer[0].p = oVector3(_R( 0.5), _R(-0.5), _R(-0.5)); // right
	buffer[1].p = oVector3(_R( 0.5), _R( 0.5), _R(-0.5));
	buffer[2].p = oVector3(_R( 0.5), _R( 0.5), _R( 0.5));
	buffer[3].p = oVector3(_R( 0.5), _R(-0.5), _R( 0.5));
	CreateBoxFace(buffer, oVector3(_R( 1), _R(0), _R(0)), length.z / textureScale.z, length.y / textureScale.y); buffer += 4;
	buffer[0].p = oVector3(_R(-0.5), _R( 0.5), _R(-0.5)); // top
	buffer[1].p = oVector3(_R(-0.5), _R( 0.5), _R( 0.5));
	buffer[2].p = oVector3(_R( 0.5), _R( 0.5), _R( 0.5));
	buffer[3].p = oVector3(_R( 0.5), _R( 0.5), _R(-0.5));
	CreateBoxFace(buffer, oVector3(_R(0), _R( 1), _R(0)), length.x / textureScale.x, length.z / textureScale.z); buffer += 4;
	buffer[0].p = oVector3(_R( 0.5), _R(-0.5), _R(-0.5)); // bottom
	buffer[1].p = oVector3(_R( 0.5), _R(-0.5), _R( 0.5));
	buffer[2].p = oVector3(_R(-0.5), _R(-0.5), _R( 0.5));
	buffer[3].p = oVector3(_R(-0.5), _R(-0.5), _R(-0.5));
	CreateBoxFace(buffer, oVector3(_R(0), _R(-1), _R(0)), length.x / textureScale.x, length.z / textureScale.z); buffer += 4;
	// scale the vertices
	for(ushort i = 0; i < numVertices; ++i)
		vertices[i].p *= length;
	// fill indices
	for(ushort i = 0, offset = 0; i < 36; offset += 4)
	{
		indices[i++] = offset; indices[i++] = offset + 1; indices[i++] = offset + 2;
		indices[i++] = offset + 2; indices[i++] = offset + 3; indices[i++] = offset;
	}
	// fill primitives
	primitives[0].type = D3DPT_TRIANGLELIST;
	primitives[0].minIndex = 0;
	primitives[0].numVertices = 24;
	primitives[0].startIndex = 0;
	primitives[0].primitiveCount = 12;
}

void Mesh::CreateCylinder(const oReal& radius, const oReal& length, const ushort& segments)
{
	// allocate
	Allocate(4 * segments + 4, 4 * segments + 6, 3);
	// fill vertices
	oArrayPtr<oReal>::Type s = new oReal[segments + 1];
	oArrayPtr<oReal>::Type c = new oReal[segments + 1];
	for(ushort i = 0; i < segments; ++i)
	{
		s[i] = oRealSin(_R(2) * oRealPi * _R(i) / _R(segments));
		c[i] = oRealCos(_R(2) * oRealPi * _R(i) / _R(segments));
	}
	s[segments] = s[0];
	c[segments] = c[0];
	MeshVertex* buffer = vertices;
	for(ushort i = 0; i < segments; ++i) // top
	{
        buffer->p = oVector3(s[i], _R(0.5), c[i]);
		buffer->n = oVector3(_R(0), _R(1), _R(0));
		buffer->t = oVector2((_R(1) +  s[i]) / _R(2), (_R(1) + c[i]) / _R(2));
		++buffer;
	}
	buffer->p = oVector3(_R(0), _R(0.5), _R(0)); // top pin
	buffer->n = oVector3(_R(0), _R(1), _R(0));
	buffer->t = oVector2(_R(0.5), _R(0.5));
	++buffer;
	for(ushort i = 0; i < segments; ++i) // bottom
	{
        buffer->p = oVector3(s[i], _R(-0.5), c[i]);
		buffer->n = oVector3(_R(0), _R(-1), _R(0));
		buffer->t = oVector2((_R(1) +  s[i]) / _R(2), (_R(1) + c[i]) / _R(2));
		++buffer;
	}
	buffer->p = oVector3(_R(0), _R(-0.5), _R(0)); // bottom pin
	buffer->n = oVector3(_R(0), _R(-1), _R(0));
	buffer->t = oVector2(_R(0.5), _R(0.5));
	++buffer;
	for(ushort i = 0; i <= segments; ++i) // sides
	{
        buffer->p = oVector3(s[i], _R(0.5), c[i]);
		buffer->n = oVector3(s[i], _R(0), c[i]);
		buffer->t = oVector2((_R(1) +  s[i]) / _R(2), (_R(1) + c[i]) / _R(2));
		//buffer->t = oVector2(_R(i) / _R(segments), _R(1));
		++buffer;
        buffer->p = oVector3(s[i], _R(-0.5), c[i]);
		buffer->n = oVector3(s[i], _R(0), c[i]);
		buffer->t = oVector2((_R(1) +  s[i]) / _R(2), (_R(1) + c[i]) / _R(2));
		//buffer->t = oVector2(_R(i) / _R(segments), _R(0));
		++buffer;
	}
	// scale
	oVector3 scale(radius, length, radius);
	for(ushort i = 0; i < numVertices; ++i)
		vertices[i].p *= scale;
	// fill indices
	ushort* ind = indices;
	*ind++ = segments; // top
	for(ushort i = 0; i < segments; ++i)
		*ind++ = i;
	*ind++ = 0;
	*ind++ = 2 * segments + 1; // bottom
	for(ushort i = 2 * segments; i > segments; --i)
		*ind++ = i;
	*ind++ = 2 * segments;
	for(ushort i = 2 * segments + 2; i < 4 * segments + 4; ++i) // sides
		*ind++ = i;
	// fill primitives
	primitives[0].type = D3DPT_TRIANGLEFAN; // top fan
	primitives[0].minIndex = 0;
	primitives[0].numVertices = segments + 1;
	primitives[0].startIndex = 0;
	primitives[0].primitiveCount = segments;
	primitives[1].type = D3DPT_TRIANGLEFAN; // bottom fan
	primitives[1].minIndex = segments + 1;
	primitives[1].numVertices = segments + 1;
	primitives[1].startIndex = segments + 2;
	primitives[1].primitiveCount = segments;
	primitives[2].type = D3DPT_TRIANGLESTRIP;
	primitives[2].minIndex = 2 * segments + 2;
	primitives[2].numVertices = 2 * segments + 2;
	primitives[2].startIndex = 2 * segments + 4;
	primitives[2].primitiveCount = 2 * segments;
}

void Mesh::CreateGroundPlane(const oVector2& extents)
{
	// allocate
	Allocate(4, 4, 1);
	// fill vertices
	static const oReal textureScale = _R(1);
	MeshVertex* buffer = vertices;
	buffer[0].p = oVector3(-extents.x, _R(0), -extents.y);
	buffer[1].p = oVector3(-extents.x, _R(0),  extents.y);
	buffer[2].p = oVector3( extents.x, _R(0), -extents.y);
	buffer[3].p = oVector3( extents.x, _R(0),  extents.y);
	buffer[0].t = oVector2(_R(0), _R(0));
	buffer[1].t = oVector2(_R(0), extents.y * textureScale);
	buffer[2].t = oVector2(extents.x * textureScale, _R(0));
	buffer[3].t = oVector2(extents.x * textureScale, extents.y * textureScale);
	buffer[0].n = buffer[1].n = buffer[2].n = buffer[3].n = oVector3(_R(0), _R(1), _R(0));
    // fill indices
	for(ushort i = 0; i < 4; ++i)
		indices[i] = i;
	// fill primitives
	primitives[0].type = D3DPT_TRIANGLESTRIP;
	primitives[0].minIndex = 0;
	primitives[0].numVertices = 4;
	primitives[0].startIndex = 0;
	primitives[0].primitiveCount = 2;
}

void Mesh::CreateSphere(const oReal& radius, const ushort& segments, const ushort& levels)
{
	if(levels < 2)
		throw _T("Mesh::CreateSphere(): Invalid number of levels requested");
	// allocate
	Allocate((levels + 1) * (segments + 1), levels * (2 * segments + 2), levels);
	// fill vertices
	MeshVertex* buffer = vertices;
    for(ushort v = 0; v <= levels; ++v)
	{
		for(ushort u = 0; u <= segments; ++u)
		{
			oReal s = oRealSin(_R(v) * oRealPi / _R(levels));
			buffer->n = oVector3(
				s * oRealSin(_R(2) * oRealPi * _R(u) / _R(segments)), 
				oRealCos(_R(v) * oRealPi / _R(levels)), 
				s * oRealCos(_R(2) * oRealPi * _R(u) / _R(segments))
			);
            buffer->p = radius * buffer->n;
			buffer->t = oVector2(_R(segments - u) / _R(segments), _R(v) / _R(levels));
			++buffer;
		}
	}
	// fill indices
	ushort* ind = indices;
	for(ushort v = 0; v < levels; ++v)
	{
		ushort offset = v * (segments + 1);
		for(ushort i = 0; i <= segments; ++i)
		{
			*ind++ = offset + i;
			*ind++ = offset + (segments + 1) + i;
		}
	}
	// fill primitives
	for(ushort v = 0; v < levels; ++v)
	{
		primitives[v].type = D3DPT_TRIANGLESTRIP;
		primitives[v].minIndex = v * (segments + 1);
		primitives[v].numVertices = 2 * (segments + 1);
		primitives[v].startIndex = v * (2 * segments + 2);
		primitives[v].primitiveCount = 2 * segments;
	}
}

void Mesh::GetVertices(MeshVertex* const& buffer, const oMatrix4x4& meshTransform, const oMatrix4x4& normalTransform) const throw()
{
	for(ushort i = 0; i < numVertices; ++i)
	{
		buffer[i].p = meshTransform * vertices[i].p;
		buffer[i].n = oTransformAs3x3(normalTransform, vertices[i].n);
		buffer[i].t = vertices[i].t;
	}
}

void Mesh::GetIndices(ushort* const& buffer, const ushort& offset) const throw()
{
	for(ushort i = 0; i < numIndices; ++i)
		buffer[i] = indices[i] + offset;
}

void Mesh::DrawIndexedPrimitive(IDirect3DDevice8* const& device, const ulong& vertexOffset, const ulong& indexOffset) const throw()
{
	for(ushort i = 0; i < numPrimitives; ++i)
	{
		device->DrawIndexedPrimitive(
			primitives[i].type, 
			primitives[i].minIndex + vertexOffset, primitives[i].numVertices, 
			primitives[i].startIndex + indexOffset, primitives[i].primitiveCount
		);
	}
}

GeomMesh::GeomMesh() : geom(0), mesh(0), vertexOffset(0), indexOffset(0) {}

GeomMesh::GeomMesh(const dGeomID& geom_, const Mesh* const& mesh_) : vertexOffset(0), indexOffset(0)
{
	Create(geom_, mesh_);
}

void GeomMesh::Create(const dGeomID& geom_, const Mesh* const& mesh_)
{
	geom = geom_;
	mesh = mesh_;
}

void GeomMesh::DrawIndexedPrimitive(IDirect3DDevice8* const& device)
{
	mesh->DrawIndexedPrimitive(device, vertexOffset, indexOffset);
}
