#ifndef __FREEFALL_ODEHELPERS__
#define __FREEFALL_ODEHELPERS__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

template<typename ID, class Deleter>
class SmartID
{
protected:
	ID id;

public:
	SmartID() throw() : id(0) {}
	~SmartID() throw() { Free(); }

	SmartID(const ID& id_) throw() : id(id_) {}
	void operator=(const ID& id_) throw()
	{
		Free();
		id = id_;
	}

	void Free() throw()
	{
		if(id)
		{
			Deleter::Free(id);
			id = 0;
		}
	}

	operator const ID&() const throw() { return id; }
};

#define ID_DELETER(name, function) template<typename ID>									\
	class name##IDDeleter { public: inline static void Free(ID& id) throw() { function } }; \
	typedef SmartID<d##name##ID, name##IDDeleter<d##name##ID> > Smart##name##ID;

ID_DELETER(World, dWorldDestroy(id);)
ID_DELETER(Space, dSpaceDestroy(id);)
ID_DELETER(Geom, dGeomDestroy(id);)
ID_DELETER(Body, dBodyDestroy(id);)
ID_DELETER(Joint, dJointDestroy(id);)
ID_DELETER(JointGroup, dJointGroupDestroy(id);)

#endif // ndef __FREEFALL_ODEHELPERS__