#pragma once

class oRenderWindow
{
protected:
	HWND handle;
	static bool active;
	static bool sizemoved;

	static LRESULT CALLBACK RenderWindowProc(HWND window, UINT msg, WPARAM wParam, LPARAM lParam) throw();

public:
	oRenderWindow() throw();
	~oRenderWindow() throw();

	void Create(const HINSTANCE& instance, const _TCHAR* const& title, const long& width, const long& height);
	
	static bool IsActive() throw() { return active; }
	static bool WasSizeMoved() throw() { bool temp = sizemoved; sizemoved = false; return temp; }
	operator const HWND&() const throw() { return handle; }
};

class oGraphics
{
protected:
	oComPtr<IDirect3D8>::Type direct3d;
	oComPtr<IDirect3DDevice8>::Type device;
	oRenderWindow window;

	ulong vertexUsage;
	bool usingHardwareVertexShaders;
	bool usingPixelShaders;

	long targetWidth, targetHeight;
	oReal targetAspectRatio;
	D3DFORMAT renderTargetFormat;
	D3DFORMAT depthStencilFormat;

	LARGE_INTEGER previousFrameCounter, thisFrameCounter;
	oReal previousFrameTime, thisFrameTime;
	oReal performanceUnit;
	static bool thisFramePaused;

public:
	oGraphics();

	void SetRenderTargetFormat(const D3DFORMAT& renderTargetFormat_) throw();
	void SetDepthStencilFormat(const D3DFORMAT& depthStencilFormat_) throw();
	void CreateFromLoader(oGraphicsLoader const& loader, const HINSTANCE& instance, const _TCHAR* const& title, const bool& fpuPreserve);

	static bool PurgeMessageQueue() throw();
	void GetPerformanceQueryResolution();
	oReal GetFrameTime() throw();

	operator IDirect3DDevice8* const&() { return device; }
	IDirect3DDevice8* const& operator->() { return device; }

	IDirect3D8* GetDirect3D() const throw() { return direct3d; }

	const HWND& GetTargetWindow() const throw() { return window; }
	ushort GetTargetWidth() const throw() { return static_cast<ushort>(targetWidth); }
	ushort GetTargetHeight() const throw() { return static_cast<ushort>(targetHeight); }
	const oReal& GetTargetAspectRatio() const throw() { return targetAspectRatio; }

	const D3DFORMAT& GetTargetFormat() const throw() { return renderTargetFormat; }
	const D3DFORMAT& GetDepthStencilFormat() const throw() { return depthStencilFormat; }

	template<class V, class Traits>
	void CreateVertexBuffer(oVertexBuffer<V, Traits>& buffer, const ulong& vertices)
	{
		// just create it
		HRESULT hr = device->CreateVertexBuffer(
			vertices * sizeof(V), 
			vertexUsage | buffer.usageTraits, 
			0, 
			D3DPOOL_DEFAULT, 
			&buffer
		);
		if(FAILED(hr))
			throw _T("oGraphics::CreateVertexBuffer(): CreateVertexBuffer failed");
	}
	template<class Traits>
	void CreateIndexBuffer(oIndexBuffer<ushort, Traits>& buffer, const ulong& indices)
	{
		// just create it (TM)
		HRESULT hr = device->CreateIndexBuffer(
			indices * sizeof(ushort), 
			vertexUsage | buffer.usageTraits, 
			D3DFMT_INDEX16, 
			D3DPOOL_DEFAULT, 
			&buffer
		);
		if(FAILED(hr))
			throw _T("oGraphics::CreateIndexBuffer(): CreateIndexBuffer failed");
	}
	void CreateVertexShader(oVertexShader& shader, ulong* const& decl, ulong* const& body);
	void CreatePixelShader(oPixelShader& shader, ulong* const& function);

	IDirect3DTexture8* LoadTexture(const _TCHAR* const& fileName, const ulong& mipLevels, const ulong& usage, const D3DFORMAT& format, const D3DPOOL& pool);
	
	template<class V, class Traits>
	void SetStreamSource(const ulong& num, const oVertexBuffer<V, Traits>& buffer)
	{
		device->SetStreamSource(num, buffer, sizeof(V));
	}
	void ReleaseStreamSource(const ulong& num);
};

typedef oComPtr<IDirect3DTexture8>::Type oTexture;
typedef oComPtr<IDirect3DCubeTexture8>::Type oCubeTexture;
typedef oComPtr<IDirect3DVolumeTexture8>::Type oVolumeTexture;
