#include "openengine.h"

oInput::oInput() throw()
{
	memset(keyboardData, 0, 256);
}

oInput::~oInput() throw()
{
	if(keyboard)
		keyboard->Unacquire();
}
	
void oInput::Create(const HINSTANCE& instance, const HWND& window)
{
	// construct the base DirectInput8 device
	DirectInput8Create(
		instance, 
		DIRECTINPUT_VERSION, 
		IID_IDirectInput8, 
		reinterpret_cast<void**>(&dinput), 
		0
	);
	if(!dinput)
		throw _T("oInput::Create(): DirectInput8Create failed");

	// create a keyboard device
	HRESULT hr = dinput->CreateDevice(GUID_SysKeyboard, &keyboard, 0);
	if(FAILED(hr))
		throw _T("oInput::Create(): CreateDevice failed");

	// set the keyboard cooperative level
	hr = keyboard->SetCooperativeLevel(window, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
	if(FAILED(hr))
		throw _T("oInput::Create(): SetCooperativeLevel failed");

	//set the keyboard data format
	hr = keyboard->SetDataFormat(&c_dfDIKeyboard);
	if(FAILED(hr))
		throw _T("oInput::Create(): SetDataFormat failed");
	
	// attempt to aquire the keyboard
	keyboard->Acquire();
}

void oInput::GetImmediateKeyboardData() throw()
{
	// get the immediate keyboard data
	HRESULT hr = keyboard->GetDeviceState(256, keyboardData);
	if(FAILED(hr))
		ZeroMemory(keyboardData, 256);
	// attempt to reacquire the device if it has been lost
	if(hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
		keyboard->Acquire();
}

bool oInput::IsKeyDown(uchar key) const throw()
{
	return ((keyboardData[key] & 0x80) != 0);
}
