#include "openengine.h"

static ushort oQuickHullFindPeak(const oVector2* const& points, std::vector<ushort>& indices, const oPlane2& line) throw()
{
	// make the peak the first element
	std::vector<ushort>::iterator peak = indices.begin();
	oReal d = line.PointHeight(points[*peak]), t;
	// check the rest to get the true peak
	std::vector<ushort>::iterator it = peak;
	for(++it; it != indices.end(); ++it)
	{
		t = line.PointHeight(points[*it]);
		if(t > d)
		{
			peak = it;
			d = t;
		}
	}
	// remove and return it
	const ushort temp = *peak;
    *peak = indices.back();
	indices.pop_back();
	return temp;
}

static void oQuickHullPartition(const oVector2* const& points, const std::vector<ushort>& indices, const oReal& tol2,
                                const ushort& left, const ushort& peak, const ushort& right, ushort*& hull) throw()
{
	// partition
	std::vector<ushort> l, r;
	oPlane2 a(points[left], points[peak]), b(points[peak], points[right]);
	for(std::vector<ushort>::const_iterator it = indices.begin(); it != indices.end(); ++it)
	{
		if(a.PointAboveTol2(points[*it], tol2))
			l.push_back(*it);
		else if(b.PointAboveTol2(points[*it], tol2))
			r.push_back(*it);
	}

	// recurse
	if(!l.empty())
	{
		const ushort leftPeak = oQuickHullFindPeak(points, l, a);
		oQuickHullPartition(points, l, tol2, left, leftPeak, peak, hull);
	}
	*hull++ = peak;
	if(!r.empty())
	{
		const ushort rightPeak = oQuickHullFindPeak(points, r, b);
		oQuickHullPartition(points, r, tol2, peak, rightPeak, right, hull);
	}
}

ushort oQuickHull(const oVector2* const& points, const ulong& count, const oReal& tol2, ushort* const& hull) throw()
{
	// take the endpoints of a maximal internal line as the first 2 extreme points
	ushort a = 0, b = 0;
	oReal d = tol2, t;
	for(ushort i = 0; i < count;)
	{
		if(i != a && i != b)
		{
			if((t = (points[i] - points[b]).NormSquared()) > d)
			{
				d = t + tol2; a = i; i = 0;
			}
			else if((t = (points[i] - points[a]).NormSquared()) > d)
			{
				d = t + tol2; b = i; i = 0;
			}
			else
				++i;
		}
		else 
			++i;
	}
	// check for validity
	if(a == b)
		return 0;

	// recursively build the convex hull
    std::vector<ushort> indices(count - 2);
    for(ushort i = 0; i < count; ++i)
	{
		if(i != a && i != b)
			indices.push_back(i);
	}
	ushort* hullEnd = hull;
	*hullEnd++ = a;
    oQuickHullPartition(points, indices, tol2, a, b, a, hullEnd);

	// compute the size of the generated hull and return it
	return static_cast<ushort>((reinterpret_cast<size_t>(hullEnd) - reinterpret_cast<size_t>(hull)) / sizeof(ushort));
}