#pragma once

#define DELETER(name, function) \
	template<class T> \
class o##name##Deleter { public: inline static void Free(T*& p) throw() { function } };

DELETER(Operator, delete p;)
DELETER(Array, delete[] p;)
DELETER(Com, p->Release();)
DELETER(Free, free(static_cast<void*>(p));)
DELETER(Close, fclose(p);)
DELETER(Null, ;)

template<class T, class Deleter = oOperatorDeleter<T> >
class oSmartPtr
{
protected:
	T* p;

public:
	oSmartPtr() throw() : p(0) {}
	
	~oSmartPtr() throw()
	{
		Free();
	}

	void Free() throw()
	{
		if(p)
		{
			Deleter::Free(p);
			p = 0;
		}
	}

	oSmartPtr(T* const& p_) throw() : p(p_) {}

	T* const& operator=(T* const& p_) throw()
	{
		Free();
		p = p_;
		return p;
	}

	T* Detach() throw()
	{
		T* temp = p;
		p = 0;
		return temp;
	}

	operator T* const&() const throw()
	{
		return p;
	}
	T* const& operator->() const
	{
		if(p == 0)
			throw _T("oSmartPtr::operator->(): Access violation");
		return p;
	}
	T** operator&()
	{
		return &p;
	}
};

template<class T>
class oHeapPtr { public: typedef oSmartPtr<T, oFreeDeleter<T> > Type; };

template<class T>
void oAlloc(oHeapPtr<T>::Type& ptr, ulong elements)
{
	// check for zero
	if(!elements)
	{
		ptr.Free();
		return;
	}
	// reallocate
	T* temp = static_cast<T*>(realloc(ptr, elements * sizeof(T)));
	if(!temp)
		throw _T("HeapAlloc(): realloc failed");
	ptr.Detach();
	ptr = temp;
}

template<class T>
class oComPtr { public: typedef oSmartPtr<T, oComDeleter<T> > Type; };

template<class T>
class oArrayPtr { public: typedef oSmartPtr<T, oArrayDeleter<T> > Type; };

typedef oSmartPtr<FILE, oCloseDeleter<FILE> > oFilePtr;

template<class P, class Deleter = oNullDeleter<P> >
class oStack
{
protected:
	oHeapPtr<P>::Type stack;
	ulong capacity, size;

	inline void Alloc(const ulong& capacity_)
	{
		oAlloc(stack, capacity_);
		capacity = capacity_;
	}

public:
	oStack() : capacity(0), size(0) {}
	~oStack() throw()
	{
		while(size)
			Deleter::Free(stack[--size]);
	}

	void Push(const P& p)
	{
		if(size == capacity)
			Alloc((size < 8) ? 8 : (size * 3 / 2));
		stack[size++] = p;
	}
	const P& Peek()
	{
		if(!size)
			throw _T("oStack::Peek(): Empty stack");
		return stack[size - 1];
	}
	void Pop() throw()
	{
		if(size)
			Deleter::Free(stack[--size]);
	}

	const ulong& GetSize() const throw() { return size; }

	const P& operator[](const ulong& index) const
	{
		if(index >= size)
			throw _T("oStack::operator[]: index out of bounds");
		return stack[index];
	}
};