#include "openengine.h"

void oRunnable::Signal() {}

oThread::oThread() : handle(0) {}

oThread::~oThread() throw()
{
	if(handle)
	{
		// if the thread is still running then terminate it
		if(WaitForSingleObject(handle, 0) == WAIT_TIMEOUT)
			TerminateThread(handle, 1);
		// close the handle
		CloseHandle(handle);
		handle = 0;
	}
}

uint oThread::ThreadProc(void* object) throw()
{
	uint ret = reinterpret_cast<oRunnable*>(object)->Run();
	// exit the thread normally
	_endthreadex(ret);
	return ret;
}

void oThread::Start(oRunnable* const& object)
{
	if(handle)
	{
		if(WaitForSingleObject(handle, 0) == WAIT_TIMEOUT)
			throw _T("oThread::Start(): Thread already running");
		CloseHandle(handle);
		handle = 0;
	}
	// create the thread
	unsigned int threadId = 0;
	handle = reinterpret_cast<HANDLE>(_beginthreadex(
		0, 
		0, 
		oThread::ThreadProc, 
		object, 
		0, 
		&threadId
	));
	if(!handle)
		throw _T("oThread::Start(): _beginthreadex failed");
}

void oThread::Wait(const ulong& timeout) throw()
{
	if(handle)
		WaitForSingleObject(handle, timeout);
}

oEvent::oEvent() : handle(0)
{
	handle = CreateEvent(0, FALSE, FALSE, 0);
	if(!handle)
		throw _T("oEvent::oEvent(): CreateEvent failed");
}

oEvent::oEvent(const bool& manualReset) : handle(0)
{
	handle = CreateEvent(0, manualReset ? TRUE : FALSE, FALSE, 0);
	if(!handle)
		throw _T("oEvent::oEvent(): CreateEvent failed");
}

oEvent::~oEvent() throw()
{
	if(handle)
	{
		CloseHandle(handle);
		handle = 0;
	}
}

void oEvent::Signal()
{
	if(!SetEvent(handle))
		throw _T("oEvent::Signal(): SetEvent failed");
}

void oEvent::Reset()
{
	if(!ResetEvent(handle))
		throw _T("oEvent::Signal(): ResetEvent failed");
}

void oEvent::Wait(const ulong& timeout) throw()
{
	if(handle)
		WaitForSingleObject(handle, timeout);
}
